/*
 * Copyright (c) 1983,1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of California at Berkeley. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#ifndef lint
static char sccsid[] = "@(#)if.c	5.6 (Berkeley) 2/7/88 plus MULTICAST 1.2";
#endif /* not lint */

#include <sys/types.h>
#include <sys/sema.h>
#include <sys/hashing.h>
#include <sys/socket.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_types.h>
#include <netinet/in.h>
#include <netinet/in_var.h>
#include <netinet/if_ether.h>
#include <netns/ns.h>
#include <arpa/inet.h>

#include <stdio.h>
#include <signal.h>
#include <sys/param.h>
#include <ctype.h>
#include <stdarg.h>
#include <errno.h>
#include <curses.h>
#include <string.h>
#include <unistd.h>
#include <bstring.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>
#include <strings.h>

#include "netstat.h"

#define	YES	1
#define	NO	0

union ifaddr_union {
	struct ifaddr ifa;
	struct in_ifaddr in;
} ifaddr;

static	void catchalarm();
static int sidewaysintpr(unsigned int, off_t);
static  char *etherprint(u_char*);

/*
 * Print a limited-length string, and indicate when it is truncated.
 * char *pat,	must be of form "...%s..."
 */
static void
limpr(int lim, char *pat, char *str)
{
	char trimstr[80];

	if (strlen(str) > lim) {
		(void)strncpy(trimstr,str,lim-1);
		trimstr[lim-1] = '~';
		trimstr[lim] = '\0';
		(void)printf(pat,trimstr);
	} else {
		(void)printf(pat,str);
	}
}

/*
 * print interface name and MTU once
 */
static void
namepr(char *name, int mtu)
{
	if (name[0] != '\0') {
		limpr(5, "%-5s", name);
		(void)printf("%-5u ", mtu);
		name[0] = '\0';
	} else {
		(void)printf("%11s","           ");
	}
}

/*
 * Print a description of an in_ifaddr structure
 */
void
print_alias(off_t inifaddraddr, off_t ifnetaddr, struct ifnet *ifnetp)
{
	struct in_ifaddr inifaddr;
	register char *cp;
	struct sockaddr_in *sin;
	int n;

	klseek(kmem, inifaddraddr, 0);
	read(kmem, (char *)&inifaddr, sizeof(struct in_ifaddr));

	if ((inifaddr.ia_addrflags & IADDR_PRIMARY) ||
	     ((off_t)(inifaddr.ia_ifa.ifa_ifp) != ifnetaddr)) {
			return;
	}
	/*
	 * compute address in local buffer of the ifa_addr sockaddr structure
	 */
	cp = ((off_t)inifaddr.ia_ifa.ifa_addr
			- inifaddraddr
			+ (char*)&inifaddr);
	sin = (struct sockaddr_in *)cp;

	switch (sin->sin_family) {

	case AF_UNSPEC:
		printf("%23s %4s\n", "", "none");
		break;

	case AF_INET:
		if (ifnetp->if_flags & IFF_POINTOPOINT) {
			/*
			 * compute address in local buffer of ifa_dstaddr
			 * sockaddr structure which has dst addr for PPP links
			 */
			cp = ((off_t)inifaddr.ia_ifa.ifa_dstaddr
				- inifaddraddr
				+ (char*)&inifaddr);
			sin = (struct sockaddr_in *)cp;
		}
		printf("%23s %-19s\n",
		       "", routename(sin->sin_addr.s_addr, 19));
		break;

	default:
		printf("%23s af %2d: ", "", sin->sin_family);
		for (cp = (char*)sin+sizeof(*sin)-1;
		     cp >= ((struct sockaddr*)sin)->sa_data;
		     --cp)
			if (*cp != 0)
				break;
		n = cp - ((struct sockaddr*)sin)->sa_data + 1;
		cp = ((struct sockaddr*)sin)->sa_data;
		if (n <= 7)
			while (--n)
				printf("%02d.", *cp++ & 0xff);
		else
			while (--n)
				printf("%02d", *cp++ & 0xff);
		printf("%02d ", *cp & 0xff);
		putchar('\n');
		break;
	}
}

/*
 * Print a description of an in_ifaddr structure
 */
void
print_inifaddr(off_t inifaddraddr, off_t ifnetaddr,
		struct ifnet *ifnetp, char *name, int stats_printed)
{
	struct in_ifaddr inifaddr;
	register char *cp;
	struct sockaddr_in *sin;
	int n;

	klseek(kmem, inifaddraddr, 0);
	read(kmem, (char *)&inifaddr, sizeof(struct in_ifaddr));

	/*
	 * compute address in local buffer of the ifa_addr sockaddr structure
	 */
	cp = ((off_t)inifaddr.ia_ifa.ifa_addr
			- inifaddraddr
			+ (char*)&inifaddr);
	sin = (struct sockaddr_in *)cp;

	switch (sin->sin_family) {

	case AF_UNSPEC:
		namepr(name, ifnetp->if_data.ifi_mtu);
		limpr(11, "%-11s ", "none");
		limpr(15, "%-15s ", "none");
		break;

	case AF_INET:
		namepr(name, ifnetp->if_data.ifi_mtu);

		if (ifnetp->if_flags & IFF_POINTOPOINT) {
			/*
			 * compute address in local buffer of ifa_dstaddr
			 * sockaddr structure which has dst addr for PPP links
			 */
			cp = ((off_t)inifaddr.ia_ifa.ifa_dstaddr
				- inifaddraddr
				+ (char*)&inifaddr);
			sin = (struct sockaddr_in *)cp;
			limpr(11, "%-11s ", "(pt-to-pt)");
		} else {
			printf("%-11s ",
				netname(htonl(inifaddr.ia_subnet),
				inifaddr.ia_subnetmask,
				Aflag ? 11 : 20));
		}
		printf("%-15s ",
			routename(sin->sin_addr.s_addr,
				       !stats_printed ? 15 : 40));
		break;

	case AF_NS:
		namepr(name, ifnetp->if_data.ifi_mtu);
		{
		struct sockaddr_ns *sns = (struct sockaddr_ns *)sin;
		u_long net;
		char netnum[8];

		*(union ns_net *) &net = sns->sns_addr.x_net;
		sprintf(netnum, "%lxH", ntohl(net));
		upHex(netnum);
		printf("ns:%-8s ", netnum);
		printf("%-15s ", ns_phost(sns));
		}
		break;

	default:
		namepr(name, ifnetp->if_data.ifi_mtu);
		printf("af%2d: ", sin->sin_family);
		for (cp = (char*)sin+sizeof(*sin)-1;
		     cp >= ((struct sockaddr*)sin)->sa_data;
		     --cp)
			if (*cp != 0)
				break;
		n = cp - ((struct sockaddr*)sin)->sa_data + 1;
		cp = ((struct sockaddr*)sin)->sa_data;
		if (n <= 7)
			while (--n)
				printf("%02d.", *cp++ & 0xff);
		else
			while (--n)
				printf("%02d", *cp++ & 0xff);
		printf("%02d ", *cp & 0xff);
		break;
	}
}

/*
 * Print a description of either an ifaddr structure OR an in_ifaddr structure
 * depending on the
 */
void
print_linkaddr(off_t ifaddraddr,
		off_t ifnet,
		struct ifnet *ifnetp,
		char *name)
{
	union ifaddr_union ifaddr;
	register char *cp;
	struct sockaddr_in *sin;
	int n;

	klseek(kmem, ifaddraddr, 0);
	read(kmem, (char *)&ifaddr, sizeof ifaddr);

	if ((off_t)(ifaddr.in.ia_ifp) != ifnet)
		return;  
	/*
	 * compute address in local buffer of the
	 * ifa_addr sockaddr structure
	 */
	cp = ((off_t)ifaddr.ifa.ifa_addr
		      - ifaddraddr
		      + (char*)&ifaddr);
	sin = (struct sockaddr_in *)cp;

	if (sin->sin_family == AF_LINK) {
		/*
		 * compute address in local buffer of the
		 * ifa_addr sockaddr structure
		 */
		cp = (char *)LLADDR((struct sockaddr_dl*)sin);
		n = ((struct sockaddr_dl*)sin)->sdl_alen;

		if (!Aflag || n == 0)
			return;
		namepr(name, ifnetp->if_data.ifi_mtu);
		if (n != 6) {
			int m = printf("<Link>");
			while (--n >= 0)
				m += printf("%x%c",
					    *cp++ & 0xff,
					    n>0 ? '.' : ' ');
			m = 28 - m;
			while (m-- > 0)
				putchar(' ');
		} else {
			printf("<Link>%21s ", etherprint((u_char*)cp));
		}
	}
}

/*
 * Print a description of an in_multi structure
 */
void
print_multiaddr(off_t multiaddr, off_t ifnetaddr)
{
	struct in_multi inm;

	klseek(kmem, multiaddr, 0);
	read(kmem, (char*)&inm, sizeof(struct in_multi));

	if (inm.inm_ifp == (struct ifnet *)ifnetaddr) {
		printf("%23s %-19s\n", "",
		       routename(inm.inm_addr.s_addr, 19));
	}
	return;
}

/*
 * Print a description of the network interfaces.
 */
void
intpr(int interval, off_t ifnetaddr_start, off_t hashtable_info_addr)
{
	struct ifnet ifnet;
	union ifaddr_union ifaddr;
	off_t ifaddraddr, ifnetfound, ifnetaddr;
	char name[IFNAMSIZ+18+1+1];
	unsigned int hashtablesize;
	unsigned short i;
	off_t h, head, hashtable_inaddr_addr;
	struct arpcom ac;
	struct hashinfo hashtable_info;
	struct hashbucket hb_entry;
	struct hashtable hashtable_inaddr[HASHTABLE_SIZE];
	char *cp;
	int n;

	if (ifnetaddr_start == 0) {
		printf("ifnet: symbol not defined\n");
		return;
	}
	if (hashtable_info_addr == 0) {

		printf("hashtable_info: symbol not defined\n");
		return;
	}
	if (interval) {
		sidewaysintpr(interval, ifnetaddr_start);
		return;
	}

	/*
	 * Read address of first ifnet structure
	 */
	(void) klseek(kmem, ifnetaddr_start, 0);
	(void) read(kmem, (char *)&ifnetaddr, sizeof ifnetaddr);

	/* print interface name and MTU once */
	printf("%-5s%-5.5s %-11.11s %-15.15s %8.8s %5.5s %8.8s %5.5s",
      "Name", "Mtu", "Network", "Address", "Ipkts", "Ierrs", "Opkts", "Oerrs");

	if (qflag) {
		printf(" %1s %3s %4s", "q", "max", "drop");
	} else {
		printf(" %5s", "Coll");
		if (tflag)
			printf(" %s", "Time");
	}
	putchar('\n');

	while (ifnetaddr) {

		/* Read ifnet struct and name for the network interface */
		klseek(kmem, ifnetaddr, 0);
		read(kmem, (char *)&ifnet, sizeof ifnet);

		ifnetfound = ifnetaddr;
		ifnetaddr = (off_t)ifnet.if_next;
		ifaddraddr = (off_t)(ifnet.in_ifaddr);

		klseek(kmem, (off_t)ifnet.if_name, 0);
		read(kmem, name, IFNAMSIZ);
		name[IFNAMSIZ] = '\0';

		if (interface != 0 &&
		 (strcmp(name, interface) != 0 || unit != ifnet.if_unit)) {
			continue;
		}

		cp = index(name, '\0');
		cp = cp + sprintf(cp, "%d", ifnet.if_unit);

		if ((ifnet.if_flags & IFF_UP) == 0) {
			*cp++ = '*';
			*cp = '\0';
		}

		/*
		 * See if inifaddr block set in ifnet structure
		 */
		if (ifaddraddr == 0) { /* none yet */
			namepr(name, ifnet.if_data.ifi_mtu);
			printf("%-11.11s ", "none");
			printf("%-15s ", "none");
		} else {
			print_inifaddr(ifaddraddr,
				       ifnetfound, &ifnet,
				       name,
				       0); /* stats not printed yet */
		}
		printf("%8u %5u %8u %5u ",
		       ifnet.if_data.ifi_ipackets, ifnet.if_data.ifi_ierrors,
		       ifnet.if_data.ifi_opackets, ifnet.if_data.ifi_oerrors);

		if (qflag) {
			printf("%2d %2d %3d",
			       ifnet.if_snd.ifq_len,
			       ifnet.if_snd.ifq_maxlen,
			       ifnet.if_snd.ifq_drops);
		} else {
			printf("%5u", ifnet.if_data.ifi_collisions);
			if (tflag)
				printf(" %3d", ifnet.if_timer);
		}
		putchar('\n');

		/*
		 * print all internet alias addresses for this interface
		 * via enumerating the internet address hash table and looking
		 * for alias entries who match this interface.
		 */
		klseek(kmem, hashtable_info_addr, 0);
		read(kmem, (char*)&hashtable_info, sizeof(struct hashinfo));

		hashtable_inaddr_addr = (off_t)hashtable_info.hashtable;
		hashtablesize = hashtable_info.hashtablesize;
		if (hashtablesize > HASHTABLE_SIZE)
			hashtablesize = HASHTABLE_SIZE;

		klseek(kmem, hashtable_inaddr_addr, 0);
		read(kmem, (char*)hashtable_inaddr,
			hashtablesize * sizeof(struct hashtable));

		for (i=0; i < hashtablesize; i++) {

			head = (off_t)(hashtable_inaddr_addr +
				       i * sizeof(struct hashtable));
			klseek(kmem, head, 0);
			read(kmem, (char *)&hb_entry, sizeof(hb_entry));

			/*
			 * Regular C code for loop thru hash table.
			 * for (h = hashtable[i].next; h != head; h = h->next)
			 */
			for (h = (off_t)hashtable_inaddr[i].next; h != head;
				h = (off_t)hb_entry.next) {

				klseek(kmem, h, 0);
				read(kmem, (char*)&hb_entry, sizeof(hb_entry));

				if (hb_entry.flags & HTF_INET) {
					print_alias((off_t)h,
						    ifnetfound, &ifnet);
				}
			}
		}

		if (aflag) {
			/*
			 * print all internet multicast addresses for this
			 * interface by enumerating the internet address
			 * hash table and looking for multicast entries
			 * who match this interface.
			 */
			for (i=0; i < hashtablesize; i++) {

				head = (off_t)(hashtable_inaddr_addr +
					       i * sizeof(struct hashtable));
				klseek(kmem, head, 0);
				read(kmem, (char *)&hb_entry,sizeof(hb_entry));

				for (h = (off_t)hashtable_inaddr[i].next;
				     h != head;
				     h = (off_t)hb_entry.next) {

					klseek(kmem, h, 0);
					read(kmem, (char*)&hb_entry,
						sizeof(hb_entry));

					if (hb_entry.flags & HTF_MULTICAST) {
						print_multiaddr((off_t)h,
							ifnetfound);
					}
				}
			}
			/*
			 * print the link-level address for this interface
			 */
			if (ifnet.if_data.ifi_type == IFT_ETHER ||
			    ifnet.if_data.ifi_type == IFT_FDDI ||
			    ifnet.if_data.ifi_type == IFT_ISO88025) {

				klseek(kmem, ifnetfound, 0);
				read(kmem, (char *)&ac, sizeof ac);
				printf("%23s %s\n", "",
				       etherprint(ac.ac_enaddr));
			}
		}
	}
	return;
}

#define	MAXIF	128
struct	iftot {
	char	ift_name[1+IFNAMSIZ+1+3+1+1];	/* "(ifnameUNIT)*" */
	union if_addr {
		struct ifaddr ifa;
		struct in_ifaddr in;
	} ift_addr;
	u_int	ift_flags;		/* interface flags */
	u_int	ift_ip;			/* input packets */
	u_int	ift_ie;			/* input errors */
	u_int	ift_ib;			/* input bytes */
	u_int	ift_op;			/* output packets */
	u_int	ift_oe;			/* output errors */
	u_int	ift_ob;			/* output bytes */
	u_int	ift_co;			/* collisions */
	u_int	ift_len;		/* queue length */
	u_int	ift_drop;		/* dropped from queue */
} iftot[MAXIF];

/*
 * Print a running summary of interface statistics.
 * Repeat display every interval seconds, showing statistics
 * collected over that interval.  Assumes that interval is non-zero.
 * First line printed at top of screen is always cumulative.
 */
static int
sidewaysintpr(unsigned int interval, off_t off)
{
	struct ifnet ifnet;
	off_t firstifnet;
	register struct iftot *ip, *total;
	register int line;
	struct iftot *lastif, *sum, *interesting;

	klseek(kmem, off, 0);
	read(kmem, (char *)&firstifnet, sizeof (off_t));
	lastif = iftot;
	sum = iftot + MAXIF - 1;
	total = sum - 1;
	interesting = iftot;
	for (off = firstifnet, ip = iftot; off;) {
		char *cp;

		klseek(kmem, off, 0);
		read(kmem, (char *)&ifnet, sizeof ifnet);
		klseek(kmem, (off_t)ifnet.if_name, 0);
		ip->ift_name[0] = '(';
		read(kmem, ip->ift_name + 1, IFNAMSIZ);
		ip->ift_name[IFNAMSIZ+1] = '\0';
		if (interface && strcmp(ip->ift_name + 1, interface) == 0 &&
		    unit == ifnet.if_unit)
			interesting = ip;
		cp = &ip->ift_name[strlen(ip->ift_name)];
		sprintf(cp, "%d)", ifnet.if_unit);
		ip++;
		if (ip >= iftot + MAXIF - 2)
			break;
		off = (off_t) ifnet.if_next;
	}
	lastif = ip;

banner:
	printf("    input   %-6.6s    output       ", interesting->ift_name);
	if (lastif - iftot > 0)
		printf("     input  (Total)    output");
	for (ip = iftot; ip < iftot + MAXIF; ip++) {
		ip->ift_ip = 0;
		ip->ift_ie = 0;
		ip->ift_op = 0;
		ip->ift_oe = 0;
		ip->ift_co = 0;
	}
	putchar('\n');
	printf("%8.8s %5.5s %8.8s %5.5s %5.5s ",
		"packets", "errs", "packets", "errs", "colls");
	if (lastif - iftot > 0)
		printf("%8.8s %5.5s %8.8s %5.5s %5.5s ",
			"packets", "errs", "packets", "errs", "colls");
	putchar('\n');
	fflush(stdout);
	line = 0;
loop:
	sum->ift_ip = 0;
	sum->ift_ie = 0;
	sum->ift_op = 0;
	sum->ift_oe = 0;
	sum->ift_co = 0;
	for (off = firstifnet, ip = iftot; off && ip < lastif; ip++) {
		klseek(kmem, off, 0);
		read(kmem, (char *)&ifnet, sizeof ifnet);
		if (ip == interesting)
			printf("%8u %5u %8u %5u %5u ",
			       ifnet.if_data.ifi_ipackets - ip->ift_ip,
			       ifnet.if_data.ifi_ierrors - ip->ift_ie,
			       ifnet.if_data.ifi_opackets - ip->ift_op,
			       ifnet.if_data.ifi_oerrors - ip->ift_oe,
			       ifnet.if_data.ifi_collisions - ip->ift_co);
		ip->ift_ip = ifnet.if_data.ifi_ipackets;
		ip->ift_ie = ifnet.if_data.ifi_ierrors;
		ip->ift_op = ifnet.if_data.ifi_opackets;
		ip->ift_oe = ifnet.if_data.ifi_oerrors;
		ip->ift_co = ifnet.if_data.ifi_collisions;
		sum->ift_ip += ip->ift_ip;
		sum->ift_ie += ip->ift_ie;
		sum->ift_op += ip->ift_op;
		sum->ift_oe += ip->ift_oe;
		sum->ift_co += ip->ift_co;
		off = (off_t) ifnet.if_next;
	}
	if (lastif - iftot > 0)
		printf("%8u %5u %8u %5u %5u ",
			sum->ift_ip - total->ift_ip,
			sum->ift_ie - total->ift_ie,
			sum->ift_op - total->ift_op,
			sum->ift_oe - total->ift_oe,
			sum->ift_co - total->ift_co);
	*total = *sum;
	putchar('\n');
	fflush(stdout);
	line++;
	 if (interval)
		sleep(interval);
	if (line == 21)
		goto banner;
	goto loop;
	/*NOTREACHED*/
}

/*
 * Return a printable string representation of an Ethernet address.
 */
static char *
etherprint(u_char *en)
{
	static char string[18];

	sprintf(string, "%02x:%02x:%02x:%02x:%02x:%02x",
		en[0], en[1], en[2], en[3], en[4], en[5] );
	string[17] = '\0';
	return(string);
}

/*----------------------------------------------------------------------*/
#include "cdisplay.h"

static struct ifnet ifnet;
static struct iftot ziftot[MAXIF];
static struct iftot niftot[MAXIF];

void
initif(int initall)
{
	if (initall)
		bzero((char*)&iftot[0], sizeof(iftot));
	bzero((char*)&ziftot[0], sizeof(ziftot));
}

void
zeroif(void)
{
	register int i;

	for (i = 0; i < MAXIF; i++)
		ziftot[i] = iftot[i];
}

void
sprif(int y, off_t ifnetaddr)
{
	off_t ifaddraddr;
	register struct iftot *nip, *oip;
	int num_ifs;
	struct sockaddr_in *sin;
	register char *cp;
	char *netstr, *addrstr, *tmpcp;
	int n;
	char netary[64], addrary[64];
	char line[200];
	union if_addr ifaddr;

	nip = niftot;

	klseek(kmem, ifnetaddr, 0);
	read(kmem, (char *)&ifnetaddr, sizeof ifnetaddr);

	move(y,0);
	printw("%-5s%-5.5s %-16.16s %s",
	       "Name", "Mtu", "Network", "Address");
	y++;

	for (num_ifs= 0; ifnetaddr != 0; y++, nip++, num_ifs++) {

		klseek(kmem, ifnetaddr, 0);
		read(kmem, (char *)&ifnet, sizeof ifnet);

		klseek(kmem, (off_t)ifnet.if_name, 0);
		read(kmem, nip->ift_name, IFNAMSIZ);

		nip->ift_name[IFNAMSIZ] = '\0';
		ifnetaddr = (off_t)ifnet.if_next;
		cp = &nip->ift_name[strlen(nip->ift_name)];
		cp = index(nip->ift_name, '\0');
		cp = cp + sprintf(cp, "%d", ifnet.if_unit);
		if ((ifnet.if_flags&IFF_UP) == 0)
			*cp++ = '*';
		*cp = '\0';

		nip->ift_ip = ifnet.if_data.ifi_ipackets;
		nip->ift_ie = ifnet.if_data.ifi_ierrors;
		nip->ift_ib = ifnet.if_data.ifi_ibytes;
		nip->ift_op = ifnet.if_data.ifi_opackets;
		nip->ift_oe = ifnet.if_data.ifi_oerrors;
		nip->ift_ob = ifnet.if_data.ifi_obytes;
		nip->ift_co = ifnet.if_data.ifi_collisions;
		nip->ift_len = ifnet.if_snd.ifq_len;
		nip->ift_drop = ifnet.if_snd.ifq_drops;

		ifaddraddr = (off_t)(ifnet.in_ifaddr);

		if (ifaddraddr == 0) {
			netstr = "none";
			addrstr = "none";
		} else {
			klseek(kmem, ifaddraddr, 0);
			read(kmem, &ifaddr, sizeof(ifaddr));

			cp = ((off_t)ifaddr.in.ia_ifa.ifa_addr
			      - ifaddraddr
			      + (char*)&ifaddr);
			sin = (struct sockaddr_in *)cp;

			/* if nothing has changed, do not bother looking
			 * up names or displaying them.
			 */
			if (!Cfirst
			    && nip->ift_flags == ifnet.if_flags
			    && !bcmp((void *)&ifaddr.in.ia_ifa,
				     (void *)&(nip->ift_addr),
				     sizeof(nip->ift_addr)))
				continue;

			nip->ift_flags = ifnet.if_flags;
			nip->ift_addr = ifaddr;

			switch (sin->sin_family) {

			case AF_UNSPEC:
			case AF_LINK:
				netstr = "none";
				addrstr = "none";
				break;

			case AF_INET:
				if ((ifnet.if_flags & IFF_POINTOPOINT) != 0) {
					/*
					 * compute offset in ifaddr structure
					 * for the destination sockaddr
					 */
				     cp = ((off_t)ifaddr.in.ia_ifa.ifa_dstaddr
					      - ifaddraddr
					      + (char*)&ifaddr);
					sin = (struct sockaddr_in *)cp;
					netstr = "(pt-to-pt)";
				} else {
				    netstr= netname(htonl(ifaddr.in.ia_subnet),
						     ifaddr.in.ia_subnetmask,
						     16);
				}
				addrstr = routename(sin->sin_addr.s_addr,0);
				break;
			case AF_NS:
				{
				struct sockaddr_ns *sns =
					(struct sockaddr_ns *)sin;
				u_long net;
				char netnum[8];

				*(union ns_net *) &net = sns->sns_addr.x_net;
				sprintf(netnum, "%lxH", ntohl(net));
				upHex(netnum);
				netstr = netnum;
				addrstr = ns_phost(sns);
				}
				break;
			default:
				sprintf(netary, "af%2d: ", sin->sin_family);
				for (cp = (char*)sin+sizeof(*sin)-1;
				     cp >= ((struct sockaddr*)sin)->sa_data;
				     --cp)
					if (*cp != 0)
						break;
				n = cp - ((struct sockaddr*)sin)->sa_data + 1;
				cp = ((struct sockaddr*)sin)->sa_data;
				tmpcp = addrary;
				if (n <= 7) {
					while (--n)
						sprintf(tmpcp++, "%02d.",
							*cp++ & 0xff);
				} else {
					while (--n)
						sprintf(tmpcp++, "%02d",
							*cp++ & 0xff);
				}
				printf("%02d ", *cp & 0xff);
				netstr = netary;
				addrstr = addrary;
				break;
			}
		}
		if (y < BOS) {
			move(y,0);
			printw("%-5s%-5u %-16.16s %s",
			       nip->ift_name, ifnet.if_data.ifi_mtu, netstr,
			       addrstr);
		}
	}

	y++;				/* a blank line */

	if (Cfirst && y < BOS) {
		move(y,0);
		printw("%-5s %10s %6s %10s %10s %6s %10s %8s %2s %4s",
		    "Name", "Ipkts", "Ierrs", "Ibytes",
		    "Opkts", "Oerrs", "Obytes",
		    "Coll", "q", "drop");
	}
	y++;

	oip =  (cmode == DELTA) ? iftot : ziftot;
	nip = niftot;
	for (n = 0; n < num_ifs; n++) {
		if (y < BOS) {
			move(y,0);
			sprintf(line,
			   "%-5s %10u %6u %10u %10u %6u %10u %8u %2u %4u",
			   nip->ift_name,
			   nip->ift_ip - oip->ift_ip,
			   nip->ift_ie - oip->ift_ie,
			   nip->ift_ib - oip->ift_ib,
			   nip->ift_op - oip->ift_op,
			   nip->ift_oe - oip->ift_oe,
			   nip->ift_ob - oip->ift_ob,
			   nip->ift_co - oip->ift_co,
			   nip->ift_len,
			   nip->ift_drop - oip->ift_drop);
			/* do not write past edge of screen,
			 * and indicate clipping */
			if (strlen(line) > 80) {
				line[79] = '~';
				line[80] = '\0';
			}
			addstr(line);
			y++;
		}

		bcopy(nip, &iftot[n], sizeof(iftot[n]));
		oip++;
		nip++;
	}
}
