/*
 * Copyright (c) 1983,1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of California at Berkeley. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#ifndef lint
char copyright[] =
"@(#) Copyright (c) 1983 Regents of the University of California.\n\
 All rights reserved.\n";
#endif /* not lint */

#ifndef lint
static char sccsid[] = "@(#)main.c	5.11 (Berkeley) 2/7/88 plus MULTICAST 1.1";
#endif /* not lint */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <sys/sbd.h>
#include <sys/immu.h>
#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <nlist.h>
#include <stdio.h>
#include <stdarg.h>
#include <curses.h>
#include <string.h>
#include <bstring.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/protosw.h>
#include <sys/syssgi.h>
#include <sys/mbuf.h>
#include <fcntl.h>
#include <invent.h>
#include <stdlib.h>
#include <unistd.h>

#include "netstat.h"

#if (_MIPS_SZLONG == 64)
#define NLIST   nlist64
#define KERNBASE 0 /* 64-bit kernels use all of /dev/kmem */
#else
#define NLIST   nlist
#define KERNBASE 0x80000000
#endif /* if 64bit */

struct mbufconst mbufconst;
struct pteconst pteconst;
struct pageconst pageconst;

int	pagesize = 0;
int	pnumshft = 0;
#undef CLBYTES
#define CLBYTES		pagesize
#undef NBPP
#define NBPP		pagesize

/* curses.h defines nl() */
#define	nl	namelist

/*
 * Values to determine which set of network statistics to obtain
 */
struct NLIST nl[] = {
#define	N_TCB		0
	{ "tcb" },
#define	N_UDB		1
	{ "udb" },
#define	N_RAWCB		2
	{ "rawcb" },
#define	N_KPTBL		3
	{ "kptbl" },
#define	N_SYSSIZE	4
	{ "Syssize" },
#define	N_IFNET		5
	{ "ifnet" },
#define	N_HOSTS		6
	{ "hosts" },
/* #define N_		7 */
	{ "_physmem_start" },		/* unused */
#define	N_RTREE		8
	{ "rt_tables" },
#define	N_RTSTAT	9
	{ "rtstat" },
#define	N_FILE		10
	{ "file" },
#define	N_UNIXSW	11
	{ "unixsw" },
/* #define N_		12 */
	{ "_physmem_start" },		/* unused */
#define N_KERNEL_MAGIC	13
	{ "kernel_magic" },
#define N_END		14
	{ "end" },
#define	N_STRST		15
	{ "strst" },
#define	N_V		16
	{ "v" },
#define	N_NMBLOCK	17
	{ "nmblock" },
#define	N_STRINFO	18
	{ "Strinfo" },
#define N_MRTPROTO	19
	{ "ip_mrtproto" },
#define N_MRTSTAT	20
	{ "mrtstat" },
#define N_MFCTABLE	21
	{ "mfctable" },
#define N_VIFTABLE	22
	{ "viftable" },
#define N_IPFORWARD	23
	{ "ipforwarding" },
#define N_PAGECONST	24
	{ "_pageconst" },
#define N_MBUFCONST	25
	{ "_mbufconst" },
#define N_HASHINFO_INADDR 26
	{ "hashinfo_inaddr" },
#define N_HASHTABLE_INADDR 27
	{ "hashtable_inaddr" },
#define N_MEMBASE	28
	{ "_physmem_start" },
#define N_PTECONST	29
	{ "_pteconst" },
#define N_LASTENTRY	29
	"",
};

#define N_TCPSTAT	N_LASTENTRY+1
#define N_UDPSTAT	N_LASTENTRY+2
#define N_IPSTAT	N_LASTENTRY+3
#define N_ICMPSTAT	N_LASTENTRY+4
#define N_IGMPSTAT	N_LASTENTRY+5

static int syspagesize(void);
static int cpuboard(void);

#define NULLPROTOX	((struct protox *) 0)
struct protox {
	u_char	pr_index;		/* index into nlist of cb head */
	u_char	pr_sindex;		/* index into nlist of stat block */
	u_char	pr_wanted;		/* 1 if wanted, 0 otherwise */
	void	(*pr_cblocks)();	/* control blocks printing routine */
	void	(*pr_stats)();		/* statistics printing routine */
	char	*pr_name;		/* well-known name */
} protox[] = {
	{ N_TCB,	N_TCPSTAT,	1,	protopr,  tcp_stats, "tcp" },
	{ N_UDB,	N_UDPSTAT,	1,	protopr,  udp_stats, "udp" },
	{ -1,		N_IPSTAT,	1,	0,        ip_stats, "ip" },
	{ -1,		N_ICMPSTAT,	1,	0,	  icmp_stats, "icmp" },
	{ -1,		N_IGMPSTAT,	1,	0,	  igmp_stats, "igmp" },
	{ -1,		-1,		0,	0,	  0,		0 }
};

static pde_t *Sysmap;
unsigned long mem_offset;	/* for IP20 and 22, which have a hole in memory */
#define system system_name	/* work around 'system()' in stdio.h */
char	*system = "/unix";
char	*kmemf = "/dev/kmem";

int	kmem;
int	kflag;
int	Aflag;
int	Mflag;
int	Cflag;
int	aflag;
int	hflag;
int	iflag;
#ifdef sgi
int	lflag;
#endif
int	mflag;
int	nflag;
int	pflag;
int	qflag;
int	rflag;
int	sflag;
int	tflag;
int	interval = 0;
char	*interface;
int	unit;
char	usage[] = "[ -AMCailqmnrstu ] [-f family] [-p proto] [-I interface] [ interval ] [ system ] [ core ]";

#ifdef sgi
mac_label  plabel;
#endif

int	af = AF_UNSPEC;

int ipforwarding;		/* let ipstats show state of flag */

off_t klseek(int fd, off_t base, int off);

static void chkiface(char *);

extern	off_t lseek();

static void disp(void);
short havemac = 0;	/* kernel supports Mandatory Access Control */
short havecipso = 0;	/* kernel supports IP security opts */

main(argc, argv)
	int argc;
	char *argv[];
{
	char *cp, *name;
	register struct protoent *p;
	register struct protox *tp;	/* for printing cblocks & stats */
	struct protox *name2protox();	/* for -p */
	off_t kern_end;

	havemac = (sysconf(_SC_MAC) == 1);
	havecipso = (sysconf(_SC_IP_SECOPTS) == 1);
	
	syssgi(SGI_CONST, SGICONST_PAGESZ, &pageconst, sizeof(pageconst), 0);
	syssgi(SGI_CONST, SGICONST_MBUF, &mbufconst, sizeof(mbufconst), 0);
	name = argv[0];
	argc--, argv++;
	while (argc > 0 && **argv == '-') {
		for (cp = &argv[0][1]; *cp; cp++)
		switch(*cp) {

		case 'A':
			Aflag++;
			break;

		case 'M':
			Mflag++;
			break;
		case 'C':
			Cflag++;
			break;
		case 'a':
			aflag++;
			break;

		case 'h':
			hflag++;
			break;

		case 'i':
			iflag++;
			break;

		case 'l':
			if (havecipso)
				lflag++;
			break;

		case 'm':
			mflag++;
			break;

		case 'n':
			nflag++;
			break;
		case 'q':
			qflag++;
			break;
		case 'r':
			rflag++;
			break;

		case 's':
			sflag++;
			break;

		case 't':
			tflag++;
			break;

		case 'u':
			af = AF_UNIX;
			break;

		case 'p':
			argv++;
			argc--;
			if (argc == 0)
				goto use;
			if ((tp = name2protox(*argv)) == NULLPROTOX) {
				fprintf(stderr, "%s: unknown or uninstrumented protocol\n",
					*argv);
				exit(10);
			}
			pflag++;
			break;

		case 'f':
			argv++;
			argc--;
			if (argc == 0) {
				fprintf(stderr, "address family missing\n");
				fprintf(stderr, "usage: %s %s\n", name, usage);
				exit(10);
			}
			if (strcmp(*argv, "ns") == 0)
			    {
				fprintf(stderr,
				       "%s: address family not supported\n",
					*argv);
				exit(10);
			    }
			else if (strcmp(*argv, "inet") == 0)
				af = AF_INET;
			else if (strcmp(*argv, "unix") == 0)
				af = AF_UNIX;
			else {
				fprintf(stderr, "%s: unknown address family\n",
					*argv);
				exit(10);
			}
			break;

		case 'I':
			iflag++;
			if (*(interface = cp + 1) == 0) {
				if ((interface = argv[1]) == 0)
					break;
				argv++;
				argc--;
			}
			for (cp = interface; isalpha(*cp); cp++)
				;
			unit = atoi(cp);
			chkiface(interface);
			*cp-- = 0;
			break;

		default:
use:
			printf("usage: %s %s\n", name, usage);
			exit(1);
		}
		argv++, argc--;
	}
	if (argc > 0 && isdigit(argv[0][0])) {
		interval = atoi(argv[0]);
		if (interval <= 0)
			goto use;
		argv++, argc--;
		iflag++;
	}

	if (havemac
	    && sgi_getplabel(&plabel) < 0) {
		perror("netstat: can't get process label");
		exit(1);
	}

	if (argc > 0) {
		system = *argv;
		argv++, argc--;
	}
	NLIST(system, nl);
	if (nl[0].n_type == 0) {
		fprintf(stderr, "%s: no namelist\n", system);
		exit(1);
	}
	if (argc > 0) {
		kmemf = *argv;
		kflag++;
	}
	kmem = open(kmemf, 0);
	if (kmem < 0) {
		fprintf(stderr, "cannot open ");
		perror(kmemf);
		exit(1);
	}
	if(kflag) {
		/* really should be done only for cores; if they specify
		 * /dev/kmem explictly, this will break things... */
		mem_offset = nl[N_MEMBASE].n_value;
	}

	if (nl[N_KERNEL_MAGIC].n_type == 0
	    || nl[N_END].n_type == 0
	    || lseek(kmem, ((long)nl[N_KERNEL_MAGIC].n_value) & ~KERNBASE -
			mem_offset, 0) == -1
	    || read(kmem, &kern_end, sizeof(kern_end)) != sizeof(kern_end)
	    || kern_end != nl[N_END].n_value) {
		fprintf(stderr, "%s: namelist wrong for %s\n",
			system, kmemf);
		exit(1);
	}

	if (kflag) {
		off_t off;
		if (nl[N_PAGECONST].n_type == 0 ||
					nl[N_PTECONST].n_value == 0) {
			fprintf(stderr, "%s: namelist wrong for %s\n",
				system, kmemf);
			exit(1);
		}
		off = nl[N_PAGECONST].n_value & ~KERNBASE;
		lseek(kmem, off-mem_offset, 0);
		read(kmem, (char *)&pageconst, sizeof(pageconst));
		off = nl[N_PTECONST].n_value & ~KERNBASE;
		lseek(kmem, off-mem_offset, 0);
		read(kmem, (char *)&pteconst, sizeof(pteconst));
	}
	
	pagesize = pageconst.p_pagesz;
	pnumshft = pageconst.p_pnumshft;

	if (kflag) {
		off_t off;

		Sysmap = (pde_t *)malloc(NPGPT * pteconst.pt_pdesize);
		if (!Sysmap) {
			fprintf(stderr,
				"%s: unable to alloc %d bytes for Sysmap\n",
				NPGPT * pteconst.pt_pdesize);
			exit(1);
		}
	
		off = nl[N_KPTBL].n_value & ~KERNBASE;
		lseek(kmem, off-mem_offset, 0);
		read(kmem, (char *)Sysmap, NPGPT * sizeof(pde_t));

		/*
		 * Get the mbuf constants.
		 */
		if (nl[N_MBUFCONST].n_value == 0) {
			fprintf(stderr, "%s: namelist wrong for %s\n",
					system, kmemf);
			exit(0);
		}
		(void)klseek(kmem, nl[N_MBUFCONST].n_value, 0);
		(void)read(kmem, (char *)&mbufconst, sizeof(struct mbufconst));
	}

	if (nl[N_IPFORWARD].n_value != 0) {
		klseek(kmem, (off_t)nl[N_IPFORWARD].n_value, 0);
		read(kmem, (char *)&ipforwarding, sizeof (ipforwarding));
	}
	
	if (Cflag) {
		disp();
	}
	if (mflag) {
		mbpr(0);
		streampr((off_t)nl[N_STRST].n_value,
			 (off_t)nl[N_STRINFO].n_value);
		exit(0);
	}
	if (pflag) {
		if (tp->pr_stats)
			(*tp->pr_stats)(tp->pr_sindex, tp->pr_name);
		else
			printf("%s: no stats routine\n", tp->pr_name);
		exit(0);
	}
	if (hflag) {
		printf("IMP sockets not supported\n");
		exit(0);
	}
	/*
	 * Keep file descriptors open to avoid overhead
	 * of open/close on each call to get* routines.
	 */
	sethostent(1);
	setnetent(1);
	if (iflag) {
		intpr(interval,
		      nl[N_IFNET].n_value,
		      nl[N_HASHINFO_INADDR].n_value);
		exit(0);
	}
	if (rflag) {
		if (sflag)
			rt_stats((off_t)nl[N_RTSTAT].n_value);
		else
			routepr((off_t)nl[N_RTREE].n_value);
		exit(0);
	}
	if (Mflag) {
		if (sflag)
			mrt_stats((off_t)nl[N_MRTPROTO].n_value,
				  (off_t)nl[N_MRTSTAT].n_value);
		else
			mroutepr((off_t)nl[N_MRTPROTO].n_value,
				 (off_t)nl[N_MFCTABLE].n_value,
				 (off_t)nl[N_VIFTABLE].n_value);
		exit(0);
	}
    if (af == AF_INET || af == AF_UNSPEC) {
	setprotoent(1);
	setservent(1);
	while (p = getprotoent()) {

		for (tp = protox; tp->pr_name; tp++)
			if (strcmp(tp->pr_name, p->p_name) == 0)
				break;
		if (tp->pr_name == 0 || tp->pr_wanted == 0)
			continue;
		if (sflag) {
			if (tp->pr_stats)
				(*tp->pr_stats)(tp->pr_sindex, p->p_name);
		} else
			if (tp->pr_cblocks)
				(*tp->pr_cblocks)(nl[tp->pr_index].n_value,
					p->p_name);
	}
	endprotoent();
    }
    if ((af == AF_UNIX || af == AF_UNSPEC) && !sflag)
	    unixpr((off_t)nl[N_V].n_value, (off_t)nl[N_FILE].n_value,
		(struct protosw *)nl[N_UNIXSW].n_value);
    exit(0);
}

static int
cpuboard()
{
	static int board = 0;
	inventory_t *inv;

	if ( board )
		return board;

	setinvent();
	while (inv = getinvent())
		if ( inv->inv_class == INV_PROCESSOR &&
			inv->inv_type  == INV_CPUBOARD ) {
			board = inv->inv_state;
			break;
		}
	endinvent();

	return board;
}

static int
syspagesize()
{
	return pagesize;
}

/*
 * Seek into the kernel for a value.
 */
off_t
klseek(fd, base, off)
	int fd, off;
	off_t base;
{
	if (kflag) { /* reading from a core dump in a file */
			extern off_t ptophys4k(off_t,pde_t []);
			extern off_t ptophys4kip17(off_t,pde_t []);
			extern off_t ptophys4kip19(off_t,pde_t []);
			extern off_t ptophystfp(off_t,pde_t []);

			/* pte format is CPU board dependent */
			switch ( cpuboard() ) {
			case INV_IP17BOARD:
				base = ptophys4kip17(base,Sysmap);
				break;
			case INV_IP19BOARD:
				base = ptophys4kip19(base,Sysmap);
				break;
			default:
			case INV_IP20BOARD:
			case INV_IP22BOARD:
				base = ptophys4k(base,Sysmap);
				break;
			case INV_IP21BOARD:
				base = ptophystfp(base,Sysmap);
				break;
			case INV_IP25BOARD:
				base = ptophys10k(base,Sysmap);
				break;
			}
		/* allow for hole in IP20 and 22, nop for others */
		/* need this so tests for 0 still work, and so physical
		 * addresses don't get messed up */
		if(base != KERNBASE && base >= mem_offset)
			base -= mem_offset;
	} else {
		base &= ~KERNBASE;
	}
	return (lseek(fd, base, off));
}

char *
plural(n)
	int n;
{

	return (n != 1 ? "s" : "");
}

char *
pluraly(n)
	int n;
{
	return (n != 1 ? "ies" : "y");
}

char *
plurales(n)
	int n;
{
	return (n != 1 ? "es" : "");
}

/*
 * Find the protox for the given "well-known" name.
 */
struct protox *
knownname(name)
	char *name;
{
	struct protox *tp;

	for (tp = protox; tp->pr_name; tp++)
		if (strcmp(tp->pr_name, name) == 0)
			return(tp);
#ifndef sgi
	for (tp = nsprotox; tp->pr_name; tp++)
		if (strcmp(tp->pr_name, name) == 0)
			return(tp);
#endif
	return(NULLPROTOX);
}

/*
 * Find the protox corresponding to name.
 */
struct protox *
name2protox(name)
	char *name;
{
	struct protox *tp;
	char **alias;			/* alias from p->aliases */
	struct protoent *p;

	/*
	 * Try to find the name in the list of "well-known" names. If that
	 * fails, check if name is an alias for an Internet protocol.
	 */
	if (tp = knownname(name))
		return(tp);

	setprotoent(1);			/* make protocol lookup cheaper */
	while (p = getprotoent()) {
		/* assert: name not same as p->name */
		for (alias = p->p_aliases; *alias; alias++)
			if (strcmp(name, *alias) == 0) {
				endprotoent();
				return(knownname(p->p_name));
			}
	}
	endprotoent();
	return(NULLPROTOX);
}

#ifdef sgi
static void
chkiface(char *iface)
{
	int s;
	static struct	ifreq ifr;

	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		perror("netstat: socket");
		exit(1);
	}
	strncpy(ifr.ifr_name, iface, sizeof(ifr.ifr_name));
	if (ioctl(s, SIOCGIFFLAGS, (caddr_t)&ifr) < 0) {
		if (errno == ENXIO)
			fprintf(stderr, "%s: unknown interface\n", iface);
		else
			perror("netstat: ioctl (SIOCGIFFLAGS)");
		exit(1);
	}
	close(s);
}

/*----------------------------------------------------------------------*/
#include "cdisplay.h"

int Cfirst = 1;
enum cmode cmode = DELTA;

void
quit(int stat)
{
	if (Cflag) {
		move(BOS, 0);
		clrtoeol();
		refresh();
		(void)endwin();
	}
	exit(stat);
}

void
sig_winch(int sig)
{
	if (Cflag) {
		(void)endwin();
		initscr();
		Cfirst = 1;
	}

	/* re-arm in case we have System V signals */
	(void)signal(sig, sig_winch);
}

void
fail(char *fmt, ...)
{
	va_list args;

	if (Cflag) {
		move(BOS, 0);
		clrtoeol();
		refresh();
		endwin();
	}
	va_start(args, fmt);
	fprintf(stderr, "netstat: ");
	vfprintf(stderr, fmt, args);
	va_end(args);
	exit(1);
}


#define LABEL_LEN 14
#define NUM_LEN	  12
#define P(lab,fmt,val) \
	{ \
	    if (Cfirst) { move(y,x); printw("%-.*s", LABEL_LEN, lab); } \
	    move(y,x); \
	    printw(fmt,(val)); \
	    y++; \
	}
#define PL(lab) \
	{ \
	    if (Cfirst) { move(y,x); printw("%-s",lab); } \
	    y++; \
	}


static void
spr_if(int cmd, register int y, char *tmb)
{
	move(0,0);
	printw("  %d: Interface Statistics -- %s", cmd, tmb); y++;
	sprif(y, nl[N_IFNET].n_value);
}

static void
spr_mb(int cmd, register int y, char *tmb)
{
	move(0,0);
	printw("  %d: Mbuf Usage -- %s", cmd, tmb); y++;
	sprmb(y);
}


static void
spr_stream(int cmd, register int y, char *tmb)
{
	move(0,0);
	printw("  %d: Stream Buffer Usage -- %s", cmd, tmb); y++;
	sprstrm(y, (off_t)nl[N_STRST].n_value, (off_t)nl[N_STRINFO].n_value);
}

static void
spr_ip(int cmd, register int y, char *tmb)
{
	move(0,0);
	printw("  %d: IP Statistics -- %s", cmd, tmb); y++;
	/* 2nd parm old scheme was nl[N_IPSTAT].n_value) */
	sprip(y, 0);
}

static void
spr_icmp(int cmd, register int y, char *tmb)
{
	move(0,0);
	printw("  %d: ICMP Statistics -- %s", cmd, tmb); y++;

	/* 2nd parm old scheme was nl[N_ICMPSTAT].n_value) */
	spricmp(y, 0);
}

static void
spr_udp(int cmd, register int y, char *tmb)
{
	move(0,0);
	printw("  %d: UDP Statistics -- %s", cmd, tmb); y++;

	/* 2nd param old scheme was nl[N_UDPSTAT].n_value) */
	sprudp(y, 0);
}

static void
spr_tcp(int cmd, register int y, char *tmb)
{
	move(0,0);
	printw("  %d: TCP Statistics -- %s", cmd, tmb); y++;

	/* 2nd param old scheme was nl[N_TCPSTAT].n_value) */
	sprtcp(y, 0);
}


static void
spr_multi(int cmd, register int y, char *tmb)
{
	move(0,0);
	printw("  %d: Multicast Forwarding -- %s", cmd, tmb); y++;

	sprmulti(y, nl[N_MRTSTAT].n_value, nl[N_VIFTABLE].n_value,
		 nl[N_MFCTABLE].n_value);
}


static void
zero_if(void)
{
	sprif(1, nl[N_IFNET].n_value);
	zeroif();
}

static void
zero_mb(void)
{
	sprmb(1);
	zeromb();
}


static void
zero_stream(void)
{
	sprstrm(1, (off_t)nl[N_STRST].n_value, (off_t)nl[N_STRINFO].n_value);
	zerostream();
}

static void
zero_ip(void)
{
	/* 2nd parm old scheme was nl[N_IPSTAT].n_value */
	sprip(1, 0);
	zeroip();
}

static void
zero_icmp(void)
{
	/* 2nd parm old scheme was nl[N_ICMPSTAT].n_value */
	spricmp(1, 0);
	zeroicmp();
}

static void
zero_udp(void)
{
	/* 2nd parm old scheme was nl[N_UDSTAT].n_value) */
	sprudp(1, 0);
	zeroudp();
}

static void
zero_tcp(void)
{
	/* 2nd parm old scheme was nl[N_TCPSTAT].n_value) */
	sprtcp(1, 0);
	zerotcp();
}

static void
zero_multi(void)
{
	sprmulti(1, nl[N_MRTSTAT].n_value, nl[N_VIFTABLE].n_value,
		 nl[N_MFCTABLE].n_value);
	zeromulti();
}

#define NUMFUNCS 8
static void (*sprfuncs[NUMFUNCS])(int, int, char *) = {
    spr_if,
    spr_mb,
    spr_stream,
    spr_ip,
    spr_icmp,
    spr_tcp,
    spr_udp,
    spr_multi
};
static void (*sprinitfuncs[NUMFUNCS])(int) = {
    initif,
    initmb,
    initstream,
    initip,
    initicmp,
    inittcp,
    initudp,
    initmulti
};
static void (*sprzerofuncs[NUMFUNCS])() = {
    zero_if,
    zero_mb,
    zero_stream,
    zero_ip,
    zero_icmp,
    zero_tcp,
    zero_udp,
    zero_multi
};

const char min_cmd_num = '1';
const char max_cmd_num = '1' + NUMFUNCS - 1;
const char ctrl_l = 'L' & 037;

/*
 * Repeated periodic Curses display
 */
static void
disp(void)
{
	time_t now;
	char tmb[80];
	struct timeval wait;
	fd_set rmask;
	struct termio  tb;
	int c, n;
	int intrchar;			/* user's interrupt character */
	int suspchar;			/* job control character */
	int scmd = 0;			/* default scmd = MAC */
	struct tms tm;
	static char dtitle[] = "D: Delta/second";
	static char rtitle0[] = "R: Normal -- %b %d %T";
	static char rtitle[] = "R: Normal -- MMM DD HH:MM:SS";
	static char ztitle0[] = "Z: Delta -- %b %d %T";
	static char ztitle[]  = "Z: Delta -- MMM DD HH:MM:SS";

	if (interval == 0)
		interval = 1;		/* default to 1 sec interval */

	(void) ioctl(0, TCGETA, &tb);
	intrchar = tb.c_cc[VINTR];
	suspchar = tb.c_cc[VSUSP];
	FD_ZERO(&rmask);

	initmb(1);
	initstream(1);

	/* keep those open to avoid overhead */
	sethostent(1);
	setnetent(1);
	setprotoent(1);
	setservent(1);

	now = time(0) - times(&tm)/HZ;
	cftime(rtitle,rtitle0,&now);

	/* now that everything else is ready, leave cooked mode
	 */
	initscr();
	cbreak();
	noecho();
	keypad(stdscr, TRUE);
	leaveok(stdscr, FALSE);
	nonl();
	intrflush(stdscr,FALSE);
	move(0, 0);

	(void)signal(SIGINT, quit);
	(void)signal(SIGTERM, quit);
	(void)signal(SIGHUP, quit);
	(void)signal(SIGWINCH, sig_winch);
	(void)signal(SIGCONT, sig_winch);

	for (;;) {
		if (Cfirst) {
			clear();
			standout();
			move(BOS, 0);
			printw(
"1:Interface  2:Mbuf  3:Streams  4:IP  5:ICMP  6:TCP  7:UDP  8:MCast   DZRN:mode");
			standend();
			switch (cmode) {
			    case DELTA:
				move(0, 80-sizeof(dtitle));
				printw(dtitle);
				break;
			    case ZERO:
				move(0, 80-sizeof(ztitle));
				printw(ztitle);
				break;
			    case NORM:
				move(0, 80-sizeof(rtitle));
				printw(rtitle);
				break;
			}
		}
		now = time(0);
		cftime(tmb,"%b %e %T", &now);

		sprfuncs[scmd](scmd+1, 1, tmb);

		Cfirst = 0;
		move(0, 0);
		refresh();

		(void)gettimeofday(&wait, 0);
		wait.tv_sec = interval-1;
		wait.tv_usec = 1000000 - wait.tv_usec;
		if (wait.tv_usec >= 1000000) {
			wait.tv_sec++;
			wait.tv_usec -= 1000000;
		}
		if (wait.tv_sec < interval
		    && wait.tv_usec < 1000000/10)
			wait.tv_sec++;
		FD_SET(0, &rmask);
		n = select(1, &rmask, NULL, NULL, &wait);
		if (n < 0) {
			if (errno == EAGAIN || errno == EINTR)
				continue;
			fail("select: %s\n", strerror(errno));
			break;
		} else if (n == 1) {
			c = getch();
			if (c == intrchar || c == 'q' || c == 'Q') {
				quit(0);
				break;
			} else if (c == suspchar) {
				reset_shell_mode();
				kill(getpid(), SIGTSTP);
				reset_prog_mode();
			} else if (c == 'z' || c == 'Z') {
				now = time(0);
				cftime(ztitle,ztitle0,&now);
				for (n = 0; n < NUMFUNCS; n++)
					sprzerofuncs[n]();
				cmode = ZERO;
			} else if (c == 'r' || c == 'R') {
				for (n = 0; n < NUMFUNCS; n++)
					sprinitfuncs[n](0);
				cmode = NORM;
			} else if (c == 'd' || c == 'D') {
				cmode = DELTA;
			} else if (c == ctrl_l) {
				;
			} else if (c == 'n' || c == 'N') {
				nflag = !nflag;
			} else if ((c >= min_cmd_num) && (c <= max_cmd_num)) {
				scmd = c - '1';
			}
			Cfirst = 1;
		}
	}
}
#endif
