/*
 * Copyright (c) 1983,1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of California at Berkeley. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#ifndef lint
static char sccsid[] = "@(#)mbuf.c 5.5 (Berkeley) 2/7/88 plus MULTICAST 1.1";
#endif /* not lint */

#include <stdio.h>
#include <sys/param.h>
#include <sys/mbuf.h>
#include <sys/types.h>
#include <ctype.h>
#include <stdarg.h>
#include <errno.h>
#include <curses.h>
#include <string.h>
#include <bstring.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/sysmp.h>
#include <sys/tcpipstats.h>
#include "cdisplay.h"
#include "netstat.h"

#define	YES	1

extern int	pagesize;
#undef CLBYTES
#define CLBYTES		pagesize

struct kna kna;
struct	mbstat mbstat, zmbstat;

static struct mbtypes {
	int	mt_type;
	char	*mt_name;
} mbtypes[] = {
	{ MT_DATA,	"data" },
	{ MT_HEADER,	"packet headers" },
	{ MT_SOCKET,	"socket structures" },
	{ MT_PCB,	"protocol control blocks" },
	{ MT_RTABLE,	"routing table entries" },
	{ MT_HTABLE,	"IMP host table entries" },
	{ MT_ATABLE,	"address resolution tables" },
	{ MT_SONAME,	"socket names and addresses" },
	{ MT_SAT,	"security audit trail buffers" }, 
	{ MT_SOOPTS,	"socket options" },
	{ MT_FTABLE,	"fragment reassembly queue headers" },
	{ MT_RIGHTS,	"access rights" },
	{ MT_IFADDR,	"interface addresses" }, 
	{ MT_DN_DRBUF,	"4DDN driver buffers" }, 
	{ MT_DN_BLK,	"4DDN block allocator" }, 
/*	{ MT_DN_BD,	"4DDN board-mbuf conversions" }, (OBSOLETE) */
	{ MT_MCAST_RINFO, "multicast route info structures" }, 
	{ MT_IPMOPTS,	"internet multicast options" }, 
	{ MT_MRTABLE,	"multicast routing structures" }, 
	{ MT_SOACL,	"access control list for socket" }, 
	{ 0, 0 }
};

int nmbtypes = sizeof(mbstat.m_mtypes) / sizeof(u_int);
u_int seen[256];			/* "have we seen this type yet?" */

static u_int shown[256];

/*
 * Print mbuf statistics.
 */
void
mbpr(off_t mbaddr)
{
	register int totmem, totfree, totmbufs;
	register int i, err;
	register struct mbtypes *mp;

	if (nmbtypes > MT_MAX) {
		fprintf(stderr, "unexpected change to mbstat; check source\n");
		return;
	}
	err = sysmp(MP_SAGET, MPSA_TCPIPSTATS, &kna, sizeof(struct kna));
	if (err < 0) {
		fprintf(stderr, "mbpr: failed sysmp MP_SAGET\n");
		return;
	}
	mbstat = kna.mbstat;

	printf("%u/%u mbufs in use:\n",
		mbstat.m_mbufs - mbstat.m_mtypes[MT_FREE], mbstat.m_mbufs);
	totmbufs = 0;
	for (mp = mbtypes; mp->mt_name; mp++)
		if (mbstat.m_mtypes[mp->mt_type]) {
			seen[mp->mt_type] = YES;
			printf("\t%u mbufs allocated to %s\n",
			    mbstat.m_mtypes[mp->mt_type], mp->mt_name);
			totmbufs += mbstat.m_mtypes[mp->mt_type];
		}
	seen[MT_FREE] = YES;
	for (i = 0; i < nmbtypes; i++)
		if (!seen[i] && mbstat.m_mtypes[i]) {
			printf("\t%u mbufs allocated to <mbuf type %d>\n",
			    mbstat.m_mtypes[i], i);
			totmbufs += mbstat.m_mtypes[i];
		}
	if (totmbufs != mbstat.m_mbufs - mbstat.m_mtypes[MT_FREE])
		printf("*** %d mbufs missing ***\n",
			(mbstat.m_mbufs - mbstat.m_mtypes[MT_FREE]) - totmbufs);
	printf("%u/%u mapped pages in use\n",
		mbstat.m_clusters - mbstat.m_clfree, mbstat.m_clusters);

	totmem = mbstat.m_clusters * CLBYTES; /* all memory is in clusters */
	totfree = mbstat.m_mtypes[MT_FREE]*MSIZE;  /* cl_free not used */

	printf("%u Kbytes allocated to network (%d%% in use)\n",
		totmem / 1024, (totmem - totfree) * 100 / totmem);
	printf("%u requests for memory denied\n", mbstat.m_drops);
	printf("%u requests for memory delayed\n", mbstat.m_wait);
	printf("%u calls to protocol drain routines\n", mbstat.m_drain);
}

void
initmb(int initall)
{
	if (initall)
		bzero((char*)&mbstat, sizeof(mbstat));
	bzero(&zmbstat, sizeof(zmbstat));
}

void
zeromb(void)
{
	zmbstat = mbstat;
}

void
sprmb(int y)
{
	register int i, err;
	register int x = 0;
	register int totmem, totfree, totmbufs;
	register struct mbtypes *mp;
	struct	mbstat tmbstat, *mbsp;
	static int oldy = 0;
	char buf[30];

	mbsp = (cmode == DELTA) ? &mbstat : &zmbstat;

	err = sysmp(MP_SAGET, MPSA_TCPIPSTATS, &kna, sizeof(struct kna));
	if (err < 0) {
		fail("sprmb: failed sysmp MP_SAGET\n");
		return;
	}
	tmbstat = kna.mbstat;

	move(y,x);
	printw("%5u out of %u mapped pages in use\t\t",
		tmbstat.m_clusters - tmbstat.m_clfree, tmbstat.m_clusters);

	totmem = tmbstat.m_clusters * CLBYTES; /* all memory is in clusters */
	totfree = tmbstat.m_mtypes[MT_FREE]*MSIZE;  /* cl_free not used */

	move(++y,x);
	printw("%10u Kbytes allocated to network (%d%% in use)\t\t",
		totmem / 1024, (totmem - totfree) * 100 / totmem);

	move(++y,x);
	printw("%10u requests for memory denied",
		tmbstat.m_drops - mbsp->m_drops);

	move(++y,x);
	printw("%10u requests for memory delayed",
		tmbstat.m_wait - mbsp->m_wait);

	move(++y,x);
	printw("%10u calls to protocol drain routines",
		tmbstat.m_drain - mbsp->m_drain);

	totmbufs = 0;
	for (i = 0; i < MT_MAX; i++) {
		totmbufs += tmbstat.m_mtypes[i];
	}
	totmbufs = tmbstat.m_mbufs - totmbufs;
	if (totmbufs) {
		sprintf(buf, "   *** %d missing ***", totmbufs);
	} else {
		buf[0] = '\t';
		buf[1] = '\t';
		buf[2] = '\t';
		buf[3] = '\0';
	}
	move(++y,x);
	printw("%5u out of %u mbufs in use%s\t",
		tmbstat.m_mbufs - tmbstat.m_mtypes[MT_FREE],
		tmbstat.m_mbufs, buf);

	bzero(seen, sizeof(seen));
	for (mp = mbtypes;  mp->mt_name && y < BOS-1; mp++) {
		seen[mp->mt_type] = YES;
		if (tmbstat.m_mtypes[mp->mt_type] == 0
		    && !shown[mp->mt_type])
			continue;
		move(++y,0);
		printw("%16u allocated to %s\t\t",
			tmbstat.m_mtypes[mp->mt_type], mp->mt_name);
		shown[mp->mt_type] = YES;
	}
	seen[MT_FREE] = YES;

	for (i = 0; i < nmbtypes && y < BOS-1; i++) {
		if (!seen[i] && (tmbstat.m_mtypes[i]
		    		 || shown[mp->mt_type])) {
			move(++y,0);
			printw("%16u allocated to type %d\t\t",
				   tmbstat.m_mtypes[i], i);
			shown[mp->mt_type] = YES;
		}
	}
	if (y < oldy)
		clrtobot();
	oldy = y;
	mbstat = tmbstat;
}
