/*
 * Print streams memory statistics
 */

#ident "$Revision: 1.15 $"


#include <stdio.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stream.h>
#include <sys/strstat.h>
#include <sys/var.h>
#include <ctype.h>
#include <stdarg.h>
#include <errno.h>
#include <curses.h>
#include <string.h>
#include <bstring.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>
#include <stdlib.h>
#include <unistd.h>

#include "cdisplay.h"
#include "netstat.h"

/*
 * NOTICE: The following header definitions were STOLEN from kernel <sys/strsubr.h>
 * Simply trying to include strstubr.h causes lots of error due to having
 * to suck in lots of other kernel internal related header files. Too much
 * confusion and LOTS of work for one simple structure definition.
 *
 * Structure to keep track of resources that have been allocated
 * for streams - an array of these are kept, one entry per
 * resource.  This is used by crash to dump the data structures.
 */

struct strinfo {
	void	*sd_head;	/* head of in-use list */
	int	sd_cnt;		/* total # allocated */
};

#define DYN_STREAM	0	/* for stream heads */
#define DYN_QUEUE	1	/* for queues */
#define DYN_MSGBLOCK	2	/* for message blocks */
#define DYN_MDBBLOCK	3	/* for mesg/data/buffer triplets */
#define DYN_LINKBLK	4	/* for mux links */
#define DYN_STREVENT	5	/* for stream event cells */
#define DYN_QBAND	6	/* for qband structures */
#define NDYNAMIC	7	/* number of different data types dynamically allocated*/

static  void avail_strcts(void);
static unsigned int nstream, nqueue, nmblock, ndblock, nlinkblk, nseevent;
static struct strstat cstrst, zstrst;
static struct strinfo streaminfo[NDYNAMIC];

static struct label {
	char 	*txt;
	alcdat	*dat;
	alcdat	*zero;
	unsigned int *v;
} labels[] = {
	{"streams",    &cstrst.stream,  &zstrst.stream, &nstream},
	{"events",     &cstrst.strevent,&zstrst.strevent, &nseevent},
	{"queues",     &cstrst.queue,   &zstrst.queue, &nqueue},
	{"link blks",  &cstrst.linkblk, &zstrst.linkblk, &nlinkblk},
	{"mdb blks",   &cstrst.mdbblock,&zstrst.mdbblock, &ndblock},
	{"msg blks",   &cstrst.msgblock,&zstrst.msgblock, &nmblock},
};

#define NUMCTRS (sizeof(labels)/sizeof(labels[0]))

/*
 * Compute the number of available stream related data structures.
 * This must be computed as this value is dynamic.
 * NOTE: The absolute value must be returned as in DELTA mode the
 * values could become negative since the 'cstrst' structure is replaced
 * by the beginning values at the start of the timing interval.
 */
void
avail_strcts()
{
	nstream = abs(streaminfo[DYN_STREAM].sd_cnt - cstrst.stream.use);
	nseevent = abs(streaminfo[DYN_STREVENT].sd_cnt - cstrst.strevent.use);
	nqueue = abs(streaminfo[DYN_QUEUE].sd_cnt - cstrst.queue.use);
	nlinkblk = abs(streaminfo[DYN_LINKBLK].sd_cnt - cstrst.linkblk.use);
	ndblock = abs(streaminfo[DYN_MDBBLOCK].sd_cnt - cstrst.mdbblock.use);
	nmblock = abs(streaminfo[DYN_MSGBLOCK].sd_cnt - cstrst.msgblock.use);
}

static void
fetchstream(off_t staddr, off_t strinfo)
{
	static int inited = 0;
	int i, streaminfo_size;

	klseek(kmem, staddr, 0);
	i = read(kmem, &cstrst, sizeof(cstrst));
	if (i < 0) {
		perror("read(strst)");
	} else if (i != sizeof(cstrst)) {
		printf("strst: bad read\n");
	}

	if (!inited) {
		inited = 1;

		klseek(kmem, strinfo, 0);
		streaminfo_size = sizeof(streaminfo);
		i = read(kmem, &streaminfo, streaminfo_size);
		if (i < 0) {
			perror("read(strinfo)");
		} else if (i != streaminfo_size) {
			printf("strinfo: bad read\n");
		}

	}
}

void
streampr(off_t staddr, off_t strinfo)
{
	struct strstat xstrst;
	int i;
	alcdat *sp;
	struct label *lp;

	if (NUMCTRS != (sizeof(xstrst)/sizeof(alcdat))) {
		printf("strst: streams have been changed.\n");
		return;
	}

	if (staddr == 0) {
		printf("strst: symbol not in namelist\n");
		return;
	}
	if (strinfo == 0) {
		printf("Strinfo: symbol not in namelist\n");
		return;
	}

	fetchstream(staddr, strinfo);

	printf("\n%-11s %8s %6s %8s %10s %12s\n",
         "Resource", "Failures", "Avail", "In Use", "Max Used", "Total Used");

	avail_strcts();
	for (i = 0, lp = &labels[0]; i < NUMCTRS; i++,lp++) {
		sp = lp->dat;
		printf("%-10s %8d %6d %8d %10d %12d\n",
		 lp->txt, sp->fail, *lp->v, sp->use, sp->max, sp->total);
	}
}

void
initstream(int initall)
{
	bzero((char*)&zstrst, sizeof(zstrst));
}

void
zerostream()
{
	zstrst = cstrst;
}

void
sprstrm(int y, off_t staddr, off_t strinfo)
{
	alcdat *sp;
	struct label *lp;
	int i;
	int x = 0;

	fetchstream(staddr, strinfo);

	if (Cfirst) {
		move(y, x);
	printw("%-11s %8s %6s %8s %10s %12s",
	 "Resource", "Failures", "Avail", "In Use", "Max Used", "Total Used");
	}
	y++;

	avail_strcts();

	for (i = 0, lp = &labels[0]; i < NUMCTRS; i++, lp++) {
		sp = lp->dat;
		if (Cfirst) {
			move(y,0);
			printw("%-10s", lp->txt);
		}
		move(y,11);
		printw("%8d %6d %8d %10d %12d",
		       sp->fail - lp->zero->fail,
		       *lp->v,
		       sp->use,
		       sp->max,
		       sp->total - lp->zero->total);
		y++;
	}

	if (cmode == DELTA)
		zstrst = cstrst;
}
