/*
 * Copyright (c) 1983,1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of California at Berkeley. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#ifndef lint
static char sccsid[] = "@(#)unix.c	5.5 (Berkeley) 2/7/88";
#endif /* not lint */

/*
 * Display protocol blocks in the unix domain.
 */
#include <stdio.h>
#define _KMEMUSER	/*get definition of "struct file" and "struct socket" */
#include <sys/param.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/mbuf.h>
#include <sys/un.h>
#if (_MIPS_SZLONG == 32)
#define	ino_t	ino64_t
#endif
#include <sys/unpcb.h>
#if (_MIPS_SZLONG == 32)
#undef	ino_t
#endif

#include <sys/vnode.h>
#include <sys/vfs.h>
#include <sys/var.h>
#include <sys/fsid.h>
#include <sys/fstyp.h>
#if (_MIPS_SZLONG == 32)
#define	off_t	off64_t
#endif
#include <sys/file.h>
#if (_MIPS_SZLONG == 32)
#undef	off_t
#endif
#include <unistd.h>
#include <sys/mac_label.h>
#include <sys/so_dac.h>

#include "netstat.h"

extern	char *calloc();

void
unixpr(off_t vaddr,
       off_t fileaddr,
       struct protosw *unixsw)
{
	int count = 0;
	struct file fentry;
	register struct file *fp = &fentry;
	struct file *filep;
	struct socket sock, *so = &sock;
	int soc_fstyp;

	if (fileaddr == 0) {
		printf("file not in namelist.\n");
		return;
	}

	soc_fstyp = sysfs(GETFSIND, FSID_SOCKET);
	if (soc_fstyp < 0) {
		printf("soc_fstyp: sysfs failed.\n");
		return;
	}

	for (filep = (struct file *)fileaddr; filep != NULL;
	     filep = fp->f_next) {
		struct vnode vn;
		struct vfs   vfs;

		klseek(kmem, (off_t)filep, L_SET);
		if (read(kmem, (char *)fp, sizeof (struct file)) !=
		    sizeof (struct file)) {
			printf("File table read error (file=0x%x, count=%d).\n", filep, count++);
			return;
		}
		if (fp->f_count == 0)
			continue;
		klseek(kmem, (off_t)fp->f_vnode, L_SET);
		if (read(kmem, (char *)&vn, sizeof (vn)) != sizeof (vn))
			continue;
		klseek(kmem, (off_t)vn.v_vfsp, L_SET);
		if (read(kmem, (char *)&vfs, sizeof (vfs)) != sizeof (vfs))
			continue;
		if (vfs.vfs_fstype != soc_fstyp)
			continue;
		klseek(kmem, (off_t)vn.v_data, L_SET);
		if (read(kmem, (char *)so, sizeof (*so)) != sizeof (*so))
			continue;
		if (so->so_proto >= unixsw && so->so_proto <= unixsw + 2)
			if (so->so_pcb)
				unixdomainpr(so, vn.v_data);
	}
}

static	char *socktype[] =
    { "#0", "dgram", "stream", "cotsord", "raw", "rdm", "seqpacket" };

void
unixdomainpr(so, soaddr)
	register struct socket *so;
	caddr_t soaddr;
{
	struct unpcb unpcb, *unp = &unpcb;
	struct mbuf mbuf, *m;
	struct sockaddr_un *sa;
	static int first = 1;
#ifdef sgi
	mac_label slabel;
	struct soacl	sacl;
#endif

	klseek(kmem, (off_t)so->so_pcb, L_SET);
	if (read(kmem, (char *)unp, sizeof (*unp)) != sizeof (*unp))
		return;
	if (unp->unp_addr) {
		m = &mbuf;
		klseek(kmem, (off_t)unp->unp_addr, L_SET);
		if (read(kmem, (char *)m, sizeof (*m)) != sizeof (*m))
			m = (struct mbuf *)0;
		if(m)
			sa = mtod(m, struct sockaddr_un *);
	} else
		m = (struct mbuf *)0;
#ifdef sgi
#ifdef _SESMGR
	if (havemac) {
		(void)fetchlabel(&slabel, (off_t)so->so_label);
		if (! mac_dom(&plabel, &slabel)) 
			return;
		klseek(kmem, (off_t)so->so_acl, L_SET);
		kread(kmem, &sacl, sizeof(struct soacl));
	}
#endif
#endif
	if (first) {
		printf("Active UNIX domain sockets\n");
#ifdef sgi /* LABELLEDSOCKETS */
		if (lflag) 
			printf("%-8.8s ", "Label" );
		printf(
"%-8.8s %-6.6s %-6.6s %-6.6s %8.8s %8.8s %8.8s %8.8s %-25.25s ",
		       "Address", "Type", "Recv-Q", "Send-Q",
		       "Vnode", "Conn", "Refs", "Nextref", "Addr");
		if (lflag) 
			printf("%-8.8s %s \n", "Souid", "Soacl");
		else
		    putchar('\n');
#else
		printf(
"%-8.8s %-6.6s %-6.6s %-6.6s %8.8s %8.8s %8.8s %8.8s Addr\n",
		    "Address", "Type", "Recv-Q", "Send-Q",
		    "Vnode", "Conn", "Refs", "Nextref");
#endif /* sgi */
		first = 0;
	}
#ifdef sgi
#ifdef _SESMGR
	if (lflag) {
		labelpr(&slabel);
	} 
#endif
#endif
	printf("%8x %-6.6s %6d %6d %8x %8x %8x %8x",
	    soaddr, socktype[so->so_type], so->so_rcv.sb_cc, so->so_snd.sb_cc,
	    unp->unp_vnode, unp->unp_conn,
	    unp->unp_refs, unp->unp_nextref);
	if (m)
#ifdef sgi
		printf(" %-25.*s ", m->m_len - sizeof(sa->sun_family),
		    sa->sun_path);
	else
	    	printf("                           ");
#ifdef _SESMGR
	if (lflag) {	    
	    	souidpr(so->so_uid);
	    	soaclpr(&sacl);
	}
#endif
#else
		printf(" %.*s", m->m_len - sizeof(sa->sun_family),
		    sa->sun_path);
#endif
	putchar('\n');
}
